"""Rewrite Existing Functions"""

import os
import re
import pdb
import sys
import itertools
import subprocess
from functools import partial
from contextlib import suppress

_open = open


__all__ = list(globals())


_print, print = print, lambda *value, file=sys.stdout: _print(' '.join(map(str, value)) + '\n', end='', file=file)
infinity = itertools.count
makedirs = partial(os.makedirs, exist_ok=True)
breakpoint = lambda: pdb.set_trace()
# open = partial(open, encoding='u8')
popen = lambda cmd: subprocess.Popen(cmd, -1, None, -1, -1, -1, shell=True).stdout
listdir = lambda *paths: os.listdir(os.path.join(*paths))
findall = lambda pattern, string, flags=0: [(m.start(), m.end(), m.group()) for m in re.finditer(pattern, string, flags)]
split = lambda arr, cols: [arr[i:i+cols] for i in range(0, len(arr), cols)]


def detect(file):
    import chardet
    with _open(file, 'rb') as f:
        b = f.read()
    return chardet.detect(b)['encoding'] or 'utf-8'


def walk(paths, exts=''):
    paths = paths if isinstance(paths, (list, tuple)) else [paths]
    exts = [exts] if isinstance(exts, str) else exts
    exts = [ext.lower() for ext in exts]
    for path in paths:
        if os.path.isfile(path):
            if any(path.lower().endswith(ext) for ext in exts):
                yield path
        for root, folders, files in os.walk(path):
            for file in files:
                if any(file.lower().endswith(ext) for ext in exts):
                    yield os.path.join(root, file)


class open:
    def __init__(self, file):
        self.p = file

    def read(self):
        if not os.path.isfile(self.p):
            return None
        with suppress(UnicodeDecodeError):
            with _open(self.p, encoding='u8') as f:
                return f.read()
        with suppress(UnicodeDecodeError):
            with _open(self.p, encoding='gbk') as f:
                return f.read()
        with _open(self.p, 'rb') as f:
            return f.read()

    def write(self, data, mode='w'):
        root = os.path.dirname(self.p)
        if root and not os.path.exists(root):
            os.makedirs(root)
        if isinstance(data, bytes):
            mode += 'b'
        with _open(self.p, mode, encoding='u8') as f:
            f.write(data)

    def append(self, data):
        self.write(data, 'a')


__all__ = [k for k in globals() if k not in __all__]



"""Useful Classes"""

import ctypes
import socket
import struct
import configparser
from threading import Thread as _Thread


__all__ = list(globals())


class Config:
    def __init__(self, path, section='default', encoding='u8'):
        self._data = {}
        self._path = path
        self._section = section
        self._encoding = encoding

    def __iter__(self):
        yield from self.data

    def __len__(self):
        return len(self.data)

    def __contains__(self, name):
        return name in self.data

    def __delattr__(self, name):
        if name in self.data:
            del self.data[name]
            self.save()

    def __getattr__(self, name):
        if name in self.data:
            return self.data[name]

    def __setattr__(self, name, value):
        if name.startswith('_'):
            super().__setattr__(name, value)
        else:
            self.data[name] = value
            self.save()

    def __repr__(self):
        return '''Config('%s', '%s', '%s')%s''' % (
            self._path, self._section, self._encoding,
            ''.join('\n.%s=%s' % item for item in self.data.items()))

    @property
    def data(self):
        p = configparser.ConfigParser()
        p.read(self._path, encoding=self._encoding)
        if not p.has_section(self._section):
            p.add_section(self._section)
        self._data = dict(p.items(self._section))
        return self._data

    def save(self):
        p = configparser.ConfigParser()
        p.read_dict({self._section: self._data})
        with open(self._path, 'w', encoding=self._encoding) as f:
            p.write(f, False)


class Thread(_Thread):
    def __init__(self, target, *args, **kwargs):
        _Thread.__init__(self, target=target, args=args, kwargs=kwargs)

        self.result = None
        self.finish = False

        self.start()

    def run(self):
        self.result = self._target(*self._args, **self._kwargs)
        self.finish = True

    def stop(self):
        if not self.finish:
            ctypes.pythonapi.PyThreadState_SetAsyncExc(self.ident, ctypes.py_object(SystemExit))


class Tcp:
    def __init__(self, addr, port):
        self.host = not addr
        self.addr = addr
        self.port = port
        self.connect()

    def connect(self):
        if self.host:
            self.server = socket.socket()
            self.server.bind(('', self.port))
            self.server.listen(5)
            self.client, (addr, port) = self.server.accept()
        else:
            self.client = socket.socket()
            self.client.connect((self.addr, self.port))

    def close(self):
        self.client.close()
        if self.host:
            self.server.close()

    def send(self, data):
        if isinstance(data, str):
            data = data.encode()
        self.client.send(data)

    def recv(self, length):
        return self.client.recv(length)

    def send_with_header(self, data):
        if isinstance(data, str):
            data = data.encode()
        assert len(data) < 1 << 32  # max 4GB
        self.client.send(struct.pack('I', len(data)) + data)

    def recv_with_header(self):
        length = struct.unpack('I', self.recv(4))[0]  # max 4GB
        s = bytearray()
        while len(s) < length:
            s.extend(self.client.recv(length - len(s)))
        return bytes(s)


__all__ = [k for k in globals() if k not in __all__]


if __name__ == '__main__':
    import time

    def ThreadFunction(timeout, fun, *args, **kwargs):
        t = Thread(fun, *args, **kwargs)
        t.join(timeout)
        t.stop()
        return t.result

    def f():
        for i in range(3):
            time.sleep(1)
            print(i)
        return 9

    print(ThreadFunction(0.5, f))
    print(ThreadFunction(1.5, f))
    print(ThreadFunction(3.5, f))



"""Decorators"""

import sys
import time
import traceback
from functools import wraps
from threading import Thread


__all__ = list(globals())


def main(f):
    import inspect
    if '__main__' == inspect.currentframe().f_back.f_globals['__name__']:
        if len(sys.argv) == 1:
            f()
        for v in sys.argv[1:]:
            f(v)
    return f


def timeit(f):
    @wraps(f)
    def wrapper(*args, **kwargs):
        s = time.time()
        ret = f(*args, **kwargs)
        e = time.time()
        print('Running <%s>: %.4f' % (f.__name__, e - s))
        return ret
    return wrapper


def tracer(f):
    @wraps(f)
    def wrapper(*args, **kwargs):
        print('Enter function:', f.__name__)
        if args:
            print('args:', args)
        if kwargs:
            print('kwargs:', kwargs)
        s = time.time()
        ret = f(*args, **kwargs)
        e = time.time()
        print('Leave function:', e - s)
        return ret
    return wrapper


def protect(f):
    @wraps(f)
    def wrapper(*args, **kwargs):
        try:
            return f(*args, **kwargs)
        except Exception as e:
            msg = traceback.format_exc()
            print(msg, file=sys.stderr)
            return msg
    return wrapper


def surround(before=(), after=()):
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            # print(func.__name__) # for test
            [f() for f in before]
            ret = func(*args, **kwargs)
            [f() for f in after]
            return ret
        return wrapper
    return decorator


def hotkey(key='F12'):
    key = key.lower()
    def decorator(f):
        def press(key2):
            if key == str(key2).split('.')[-1].lower():
                f()
        def th():
            import pynput
            with pynput.keyboard.Listener(press) as kl:
                kl.join()
        Thread(target=th).start()
        return f
    return decorator


def threads(cnt):
    import threading
    counter = threading.BoundedSemaphore(cnt)
    def decorator(f):
        @wraps(f)
        def wrapper(*args, **kwargs):
            def th():
                try:
                    f(*args, **kwargs)
                except:
                    raise
                finally:
                    counter.release()
            counter.acquire()
            threading.Thread(target=th).start()
        return wrapper
    return decorator


__all__ = [k for k in globals() if k not in __all__]


if __name__ == '__main__':
    import random

    @threads(10)
    def delay():
        t = 1 + random.random() / 5
        assert t < 1.19
        time.sleep(t)
        print(t)

    while 1:
        delay()



"""C Program Interface"""

'''File: dll.dll

typedef struct Matrix {
    int height;
    int width;
    int value[16][16];
} mat;

mat* MatMul(mat m1, mat m2) {
    mat* mp = (mat*)malloc(sizeof(mat));
    int i, j, k;
    int sum;

    mp->height = m1.height;
    mp->width = m2.width;
    for(i = 0; i < m1.height; i++)
        for(j = 0; j < m2.width; j++) {
            sum = 0;
            for(k = 0; k < m1.width; k++)
                sum += m1.value[i][k] * m2.value[k][j];
            mp->value[i][j] = sum;
        }
    return mp;
}

mat* MatNew(int height, int width) {
    mat* mp = (mat*)malloc(sizeof(mat));
    int i, j;
    int sum = 0;

    mp->height = height;
    mp->width = width;
    for(i = 0; i < height; i++)
        for(j = 0; j < width; j++)
            mp->value[i][j] = sum++;
    return mp;
}

int MatSum(mat m) {
    int i, j;
    int sum = 0;
    for(i = 0; i < m.height; i++)
        for(j = 0; j < m.width; j++)
            sum += m.value[i][j];
    return sum;
}
'''

import ctypes
from ctypes import POINTER

__all__ = []


class Type:
    def __init__(self, value):
        self.value = value

    def __getitem__(self, item):
        return Type(self.value * item)


def struct(fields):
    class Struct(ctypes.Structure):
        _fields_ = fields
    return Struct


def fields(cls):
    return [(k, v.value) for k, v in cls.__dict__.items() if isinstance(v, Type)]


def wrap(func, argtype=None, restype=None):
    if argtype:
        func.argtype = argtype
    if restype:
        func.restype = POINTER(restype)
        return lambda *args: func(*args).contents
    else:
        return func


if __name__ == '__main__':
    CINT = Type(ctypes.c_int)

    class Mat:
        height = CINT
        width  = CINT
        value  = CINT[16][16]

    print(fields(Mat))
    Mat = struct(fields(Mat))
    print(Mat)

    dll = ctypes.CDLL('dll.dll')

    # - Normal style ---------------------------------------

    class Mat(ctypes.Structure):
        _fields_ = [
            ('height', ctypes.c_int),
            ('width', ctypes.c_int),
            ('value', ((ctypes.c_int * 16) * 16)),
        ]

    dll.MatMul.argtype = (Mat, Mat)
    dll.MatMul.restype = POINTER(Mat)

    m1 = Mat()
    m1.height = 3
    m1.width = 3
    for i in range(3):
        for j in range(3):
            m1.value[i][j] = i * 3 + j

    ret = dll.MatMul(m1, m1).contents
    print([row[:ret.width] for row in ret.value[:ret.height]])

    # - (int, int) -> Mat ---------------------------------------

    MatNew = wrap(dll.MatNew, restype=Mat)
    ret = MatNew(3, 4)
    print([row[:ret.width] for row in ret.value[:ret.height]])

    # - (Mat, Mat) -> Mat ---------------------------------------

    MatMul = wrap(dll.MatMul, argtype=(Mat, Mat), restype=Mat)
    m1 = Mat()
    m1.height = 3
    m1.width = 3
    for i in range(3):
        for j in range(3):
            m1.value[i][j] = i * 3 + j

    ret = MatMul(m1, m1)
    print([row[:ret.width] for row in ret.value[:ret.height]])

    # - Mat -> int ---------------------------------------

    MatSum = wrap(dll.MatSum, argtype=Mat)

    m1 = Mat()
    m1.height = 3
    m1.width = 3
    for i in range(3):
        for j in range(3):
            m1.value[i][j] = i * 3 + j

    ret = MatSum(m1)
    print(ret)



"""Document Operation"""

import os
import csv


def num2str(cell):
    if isinstance(cell, float) and cell.is_integer():
        cell = int(cell)
    return str(cell)


def str2value(s):
    try:
        v = float(s)
        if v.is_integer():
            v = int(v)
    except ValueError:
        v = {'none': None, 'null': None, '': None,
             'true': True, 'flase': False
        }.get(s.lower(), s)
    return v


__all__ = list(globals())


# ---------------------------------------------------------------------------
# Text
# ---------------------------------------------------------------------------


def read(file, encoding='u8'):
    if not encoding:
        with open(file, 'rb') as f:
            return f.read()
    try:
        with open(file, encoding=encoding) as f:
            return f.read()
    except UnicodeDecodeError:
        with open(file) as f:
            return f.read()


def write(file, data, encoding='u8'):
    root = os.path.dirname(file)
    if root and not os.path.exists(root):
        os.makedirs(root)
    if isinstance(data, str):
        with open(file, 'w', encoding=encoding) as f:
            f.write(data)
    else:
        with open(file, 'wb') as f:
            f.write(data)


def ReadIni(file, encoding='u8'):
    import configparser
    p = configparser.ConfigParser()
    p.optionxform = str  # fix opinion can't read key with upper case
    p.read(file, encoding=encoding)
    return {s: dict(p.items(s)) for s in p.sections()}


def WriteIni(file, dic, encoding='u8'):
    import configparser
    p = configparser.ConfigParser()
    p.optionxform = str  # fix opinion can't read key with upper case
    p.read_dict(dic)
    with open(file, 'w', encoding=encoding) as f:
        p.write(f, False)


def WriteTxt(file, data, encoding='utf-8-sig'):
    with open(file, 'w', encoding=encoding) as f:
        f.write('\n'.join(','.join(str(cell) for cell in row) for row in data))


def ReadCsv(file, encoding='utf-8-sig'):
    with open(file, encoding=encoding) as f:
        return list(csv.reader(f))


def WriteCsv(file, data, encoding='utf-8-sig', errors='ignore'):
    with open(file, 'w', newline='', encoding=encoding, errors=errors) as f:
        writer = csv.writer(f)
        writer.writerows(data)


# ---------------------------------------------------------------------------
# Excel
# ---------------------------------------------------------------------------


def WriteExcel(file, data, new_sheet='sheet1'):
    import xlwt
    xls = xlwt.Workbook('u8')
    sheet = xls.add_sheet(new_sheet, True)
    for r, row in enumerate(data):
        for c, cell in enumerate(row):
            sheet.write(r, c, cell)
    xls.save(file)


def OpenExcel(file):
    import xlrd
    import xlutils.filter

    rb = xlrd.open_workbook(file, formatting_info=True)

    # 参考xlutils.copy库内的用法 参考xlutils.filter内的参数定义style_list
    w = xlutils.filter.XLWTWriter()
    xlutils.filter.process(xlutils.filter.XLRDReader(rb, 'unknown.xls'), w)
    wb = w.output[0][1]
    wb.style_list = w.style_list
    wb.sheets = rb.sheets()

    # quick `style` reach, and `write` with copied style.
    wb.style = lambda n, r, c: wb.style_list[wb.sheets[n].cell_xf_index(r, c)]
    wb.swrite = lambda n, r, c, value: wb.get_sheet(n).write(r, c, value, wb.style(n, r, c))

    return wb


def MergeCell(data, merge, merge_x=True, merge_y=True, strip_x=False):
    data2 = []
    for sheet_data, sheet_merge in zip(data, merge):
        # merge cell
        for r1, r2, c1, c2 in sheet_merge:
            for r in range(r1, r2):
                for c in range(c1, c2):
                    if (not merge_x and c > c1) or (not merge_y and r > r1):
                        sheet_data[r][c] = None if strip_x else ''
                    else:
                        sheet_data[r][c] = sheet_data[r1][c1]
        # strip x
        if strip_x:
            sheet_data = [[cell for cell in row if cell is not None] for row in sheet_data]
        data2.append(sheet_data)
        # remove blanks in tail
        # for row in sheet_data:
        #     while len(row) and row[-1] == '':  # Good!
        #         row.pop()
    return data2


def ReadExcel(file, merge_x=True, merge_y=True, strip_x=False):
    import xlrd
    try:
        xls = xlrd.open_workbook(file, formatting_info=True)
    except xlrd.biffh.XLRDError:
        xls = xlrd.open_workbook(file)

    data = []
    for sheet in xls.sheets():
        sheet_name = sheet.name
        sheet_data = []
        for row in range(sheet.nrows):
            rows = [sheet_name] + sheet.row_values(row)
            sheet_data.append(list(map(num2str, rows)))
        data.append(sheet_data)

    # only ".xls" type contain merge_info
    merge = [sorted(sheet.merged_cells) for sheet in xls.sheets()]
    data2 = MergeCell(data, merge, merge_x, merge_y, strip_x)
    return data, merge


def ReadSheet(file, index=0):
    import xlrd
    xls = xlrd.open_workbook(file)
    sheet = xls.sheet_by_index(index)
    return [list(map(num2str, sheet.row_values(row))) for row in range(sheet.nrows)]


# ---------------------------------------------------------------------------
# Word
# ---------------------------------------------------------------------------


def Doc2Docx(file, overwrite=False):
    root, ext = os.path.splitext(file)
    file2 = root + '.docx.temp'
    if overwrite or not os.path.exists(file2):
        from win32com import client
        word = client.Dispatch('Word.Application')
        doc = word.Documents.Open(file)
        doc.SaveAs(file2, 16)
        doc.Close()
    return file2


def OpenDocx(file):
    import docx
    data = []
    merge = []
    doc = docx.Document(file)
    for table in doc.tables:
        cells  = table._cells  # see usage at `docx.table.Table._cells`
        cols   = table._column_count
        length = len(cells)

        # read text
        data.append([])
        for i, cell in enumerate(cells):
            r, c = divmod(i, cols)
            if c == 0:  # create new row
                data[-1].append([])
            data[-1][-1].append(cell.text)

        # find merged
        merge.append([])
        for i, cell in enumerate(cells):
            if cell in cells[:i]:  # only find first repeated cell
                continue
            for j in range(length - 1, 0, -1):  # find last repeated cell
                if cell is cells[j]:  # always can be found
                    break
            if i != j:  # first != last -> merged cells
                r1, c1 = divmod(i, cols)
                r2, c2 = divmod(j, cols)
                merge[-1].append((r1, r2 + 1, c1, c2 + 1))

    return data, merge


def ReadWordTexts(file):
    import docx
    if file.endswith('.doc'):
        file = Doc2Docx(file)
    doc = docx.Document(file)
    return [p.text for p in doc.paragraphs]


def ReadWord(file, merge_x=True, merge_y=True, strip_x=False):
    if file.endswith('.doc'):
        file = Doc2Docx(file)
    data, merge = OpenDocx(file)
    data2 = MergeCell(data, merge, merge_x, merge_y, strip_x)
    return data2


# ---------------------------------------------------------------------------
# Other
# ---------------------------------------------------------------------------


def ReadFile(file, merge_x=True, merge_y=True, strip_x=False):
    ext = os.path.splitext(file)[1]
    if ext in ('.csv', '.txt'):
        return ReadCsv(file)
    elif ext in ('.xls', '.xlsx'):
        return ReadExcel(file, merge_x, merge_y, strip_x)
    elif ext in ('.doc', '.docx'):
        return ReadWord(file, merge_x, merge_y, strip_x)


def File2Csv(file, merge_x=True, merge_y=True, strip_x=False):
    root, ext = os.path.splitext(file)
    data = ReadFile(file, merge_x, merge_y, strip_x)
    WriteCsv(data, root + '.csv')
    return data


def ReadFiles(files, merge_x=True, merge_y=True, strip_x=False):
    data = []
    for file in files:
        path, filename = os.path.split(file)
        data.extend([[path, filename] + row for row in ReadFile(file, merge_x, merge_y, strip_x)])
    return data


__all__ = [k for k in globals() if k not in __all__]


if __name__ == '__main__':
    from pprint import pprint

    # test read
    data = ReadExcel('../test.xls')
    # pprint(data)

    pprint(ReadWord('../test.docx', False, False, False))
    pprint(ReadWord('../test.docx', False, True,  False))
    pprint(ReadWord('../test.docx', True,  False, False))
    pprint(ReadWord('../test.docx', True,  True,  False))

    # test read/write with keep style
    wb = OpenExcel('../test2.xls')
    sheet2 = wb.get_sheet(0)
    for r in range(3):
        for c in range(3):
            sheet2.write(r, c, 'T%d%d' % (r, c), wb.style(0, r, c))
            # wb.swrite(0, r, c, 'V%d%d' % (r, c))

    wb.save('save2.xls')



"""Windows Dialog"""

import ctypes
from ctypes.wintypes import DWORD, WORD, LPARAM, LPVOID, HWND, HINSTANCE

__all__ = ['MessageBox', 'DirDialog', 'OpenFileDialog', 'SaveFileDialog']

MessageBox = lambda info, title='Message', style=0: ctypes.windll.user32.MessageBoxW(0, str(info), str(title), style)


LPOFNHOOKPROC = ctypes.c_void_p
LPCSTR = LPSTR = ctypes.c_wchar_p

OFN_ENABLESIZING    = 0x00800000
OFN_PATHMUSTEXIST   = 0x00000800
OFN_OVERWRITEPROMPT = 0x00000002
OFN_NOCHANGEDIR     = 0x00000008
MAX_PATH            = 1024

GetOpenFileName = ctypes.windll.comdlg32.GetOpenFileNameW
GetSaveFileName = ctypes.windll.comdlg32.GetSaveFileNameW


# Ref: https://gist.github.com/nicomgd/1097a0b7ca3715da4e71

class OPENFILENAME(ctypes.Structure):
    _fields_ = [
        ('lStructSize',       DWORD),
        ('hwndOwner',         HWND),
        ('hInstance',         HINSTANCE),
        ('lpstrFilter',       LPCSTR),
        ('lpstrCustomFilter', LPSTR),
        ('nMaxCustFilter',    DWORD),
        ('nFilterIndex',      DWORD),
        ('lpstrFile',         LPSTR),
        ('nMaxFile',          DWORD),
        ('lpstrFileTitle',    LPSTR),
        ('nMaxFileTitle',     DWORD),
        ('lpstrInitialDir',   LPCSTR),
        ('lpstrTitle',        LPCSTR),
        ('Flags',             DWORD),
        ('nFileOffset',       WORD),
        ('nFileExtension',    WORD),
        ('lpstrDefExt',       LPCSTR),
        ('lCustData',         LPARAM),
        ('lpfnHook',          LPOFNHOOKPROC),
        ('lpTemplateName',    LPCSTR),
        ('pvReserved',        LPVOID),
        ('dwReserved',        DWORD),
        ('FlagsEx',           DWORD)
    ]


def BuildOFN(title, filter, buf):
    ofn = OPENFILENAME()
    ofn.lStructSize = ctypes.sizeof(OPENFILENAME)
    ofn.lpstrTitle = title
    ofn.lpstrFile = ctypes.cast(buf, LPSTR)
    ofn.nMaxFile = MAX_PATH
    ofn.lpstrFilter = filter # 'All types (*.*)\0*.*\0'
    ofn.Flags = OFN_ENABLESIZING | OFN_PATHMUSTEXIST | OFN_OVERWRITEPROMPT | OFN_NOCHANGEDIR
    return ofn


def FileDialog(call, title, filter, path):
    filter = filter.replace('|', '\0')
    if filter and not filter.endswith('\0'):
        filter += '\0'
    buf = ctypes.create_unicode_buffer(path, MAX_PATH)
    ofn = BuildOFN(title, filter, buf)
    if call(ctypes.byref(ofn)):
        return buf[:].split('\0', 1)[0]


def OpenFileDialog(title=None, filter='', path=''):
    return FileDialog(GetOpenFileName, title, filter, path)


def SaveFileDialog(title=None, filter='', path=''):
    return FileDialog(GetSaveFileName, title, filter, path)


ole32 = ctypes.windll.ole32
shell32 = ctypes.windll.shell32

BIF_EDITBOX        = 0x0010
BIF_NEWDIALOGSTYLE = 0x0040
BIF_USENEWUI       = 0x0050

LPCTSTR = LPTSTR = ctypes.c_char_p


# Ref: https://github.com/Nolanlemahn/ProjectExist/blob/master/game/EasyDialogsWin.py
# Ref: https://github.com/apprenticeharper/DeDRM_tools/blob/master/DeDRM_plugin/askfolder_ed.py

class BROWSEINFO(ctypes.Structure):
    _fields_ = [
        ('hwndOwner',      HWND),
        ('pidlRoot',       LPVOID),
        ('pszDisplayName', LPTSTR),
        ('lpszTitle',      LPCTSTR),
        ('ulFlags',        ctypes.c_uint),
        ('lpfn',           LPVOID),
        ('lParam',         LPARAM),
        ('iImage',         ctypes.c_int)
    ]


def DirDialog(message=None):
    bi = BROWSEINFO()
    bi.pszDisplayName = ctypes.c_char_p(b'\0' * MAX_PATH)
    if message:
        bi.lpszTitle = message.encode('gbk')
    bi.ulFlags = BIF_USENEWUI

    shell32.SHBrowseForFolder.restype = LPVOID
    pidl = LPVOID(shell32.SHBrowseForFolder(ctypes.byref(bi)))
    if pidl:
        path = ctypes.c_char_p(b'\0' * MAX_PATH)
        shell32.SHGetPathFromIDList(pidl, path)
        ole32.CoTaskMemFree(pidl)
        result = path.value.decode('gbk')
        return result


if __name__ == '__main__':
    path = DirDialog()
    MessageBox(path)
    path = OpenFileDialog()
    MessageBox(path)
    path = SaveFileDialog()
    MessageBox(path)



# top module functions
all = help = None

# base on built-in library
from .cls import *
from .dec import *
from .new import *
from .line import *
from .useful import *
from .windll import *

# 3rd-library needed
from .auto import *
from .doc import *
from .gui import *
from .np import *


def all():
    import sys
    import inspect

    module_names = {}
    for module_name, module in globals().items():
        if inspect.ismodule(module) and module.__name__.count('.') == 1:
            for name in module.__all__:
                if name not in module_names:
                    module_names[name] = module_name
                elif name != '__all__':
                    print("Warning: '%s.%s' exist in '%s.%s'" % (module_name, name, module_names[name], name), file=sys.stderr)

    for fun in globals().values():
        if inspect.isfunction(fun):
            if fun.__name__ == '<lambda>':
                for k, v in inspect.currentframe().f_back.f_locals.items():
                    if v is fun:
                        fun.__name__ = k
            fun.__qualname__ = fun.__module__ + '.' + fun.__name__

    return ['all', 'help'] + list(module_names)


def help():
    import inspect
    print("\n# Lishixian Library")
    print("Contain %d functions." % len(__all__))
    print("\n## Top module")
    for k, v in globals().items():
        if not k.startswith('_'):
            if inspect.ismodule(v):
                print("\n## Module '%s'" % k)
            else:
                try:
                    argspec = str(inspect.signature(v))
                    print("- %s%s" % (k, argspec))
                except (TypeError, ValueError) as e:
                    print("- %s = %r" % (k, v))


__all__ = all()



"""GUI Functions"""

from threading import Thread


__all__ = list(globals())


def center(top):
    """tkinter set center"""
    top.update_idletasks()
    x = (top.winfo_screenwidth()  - top.winfo_reqwidth())  / 2
    y = (top.winfo_screenheight() - top.winfo_reqheight()) / 2
    top.geometry('+%d+%d'%(x, y))


def WrapBox(parent, w, label=''):
    import wx
    box = wx.StaticBoxSizer(wx.VERTICAL, parent, label)
    box.Add(w)
    return box


def GetClipboard():
    import wx
    do = wx.TextDataObject()
    if wx.TheClipboard.Open():
        ret = wx.TheClipboard.GetData(do)
        wx.TheClipboard.Close()
        if ret:
            return do.GetText()


def SetClipboard(text):
    import wx
    do = wx.TextDataObject()
    do.SetText(text)
    if wx.TheClipboard.Open():
        wx.TheClipboard.SetData(do)
        wx.TheClipboard.Close()


class Mover:
    def __init__(self, parent, widget):
        import wx
        self.p = parent
        self.dxy = (0, 0)
        widget.Bind(wx.EVT_LEFT_DOWN, self.OnLeftDown)
        parent.Bind(wx.EVT_LEFT_UP,   self.OnLeftUp)
        parent.Bind(wx.EVT_MOTION,    self.OnMouseMove)

    def OnLeftDown(self, evt):
        self.p.CaptureMouse()
        x, y = self.p.ClientToScreen(evt.GetPosition())
        x0, y0 = self.p.GetPosition()
        dx = x - x0
        dy = y - y0
        self.dxy = (dx, dy)

    def OnLeftUp(self, evt):
        if self.p.HasCapture():
            self.p.ReleaseMouse()

    def OnMouseMove(self, evt):
        if evt.Dragging() and evt.LeftIsDown():
            x, y = self.p.ClientToScreen(evt.GetPosition())
            fp = (x - self.dxy[0], y - self.dxy[1])
            self.p.Move(fp)


class EventThread(Thread):
    """
    Usage:
    widget.Bind(wx.PyEventBinder(id), lambda e: print(e.data))
    EventThread(widget, id, int, 7)
    """

    def __init__(self, parent, id, target=bool, *args, **kwargs):
        Thread.__init__(self, target=target, args=args, kwargs=kwargs)
        self.parent = parent
        self.id = id
        self.start()

    def run(self):
        import wx
        event = wx.PyEvent(0, self.id)
        event.data = self._target(*self._args, **self._kwargs)
        wx.PostEvent(self.parent, event)


__all__ = [k for k in globals() if k not in __all__]



"""Lambda Functions"""

import re
import sys
import json
import uuid
import random
import socket
import getpass
import hashlib
from threading import Thread

__all__ = list(globals())

s128 = bytes(range(128)).decode()
s127 = bytes(range(32, 127)).decode()

empty = lambda *v, **kv: None
freeze = lambda fn, *v, **kv: (lambda: fn(*v, **kv))

t = lambda arr: list(zip(*arr))

md5 = lambda b: hashlib.md5(b).hexdigest()
mask = lambda p: open(p + '.inv', 'wb').write(bytes(255 - b for b in open(p, 'rb').read()))
start = lambda func, *args, **kwargs: Thread(target=func, args=args, kwargs=kwargs).start()
create = lambda file: open(file, 'w').close()
pprint = lambda *value, file=sys.stdout: print(' '.join(map(str, value)) + '\n', end='', file=file)

dumps = lambda data: json.dumps(data, ensure_ascii=False, indent=2)

sort_kv  = lambda d, reverse=False: sorted(d.items(), key=lambda item: item[1], reverse=reverse)
sort_key = lambda d, reverse=False: sorted(d, key=d.__getitem__, reverse=reverse)
sort_num = lambda s: [(s, int(n)) for s, n in re.findall(r'(\D+)(\d+)', 'a%s0' % s)]

str2dict = lambda s: dict(re.findall(r'(.*?):(.*)', s))
tuple2item = lambda item: item if len(item) > 1 else item[0]

unique = lambda arr: sorted(set(arr), key=arr.index)

pc_ip = lambda: socket.gethostbyname(socket.gethostname())
pc_mac = lambda: '-'.join(re.findall('..', uuid.uuid1().hex[-12:].upper()))
pc_user = lambda: getpass.getuser()

randombytes = lambda n: bytes(random.randint(0, 255) for i in range(n))

join = lambda *s, sp='': sp.join(s)


__all__ = [k for k in globals() if k not in __all__]



"""Regedit"""

import winreg


__all__ = list(globals())


def GetDir(name, local=False): # Name: 'Desktop', 'SendTo', 'Programs', 'Startup'
    path = r'Software\Microsoft\Windows\CurrentVersion\Explorer\Shell Folders'
    if local:
        name = 'Common ' + name
        key = winreg.OpenKey(winreg.HKEY_LOCAL_MACHINE, path)
    else:
        key = winreg.OpenKey(winreg.HKEY_CURRENT_USER, path)
    return winreg.QueryValueEx(key, name)[0]


def DeleteRecu(key):
    for i in range(winreg.QueryInfoKey(key)[0]):
        name = winreg.EnumKey(key, i)
        DeleteRecu(winreg.OpenKey(key, name))
        winreg.DeleteKey(key, name)


def DeleteKey(key, name=''):
    root, p = key.split('\\', 1)
    key2 = winreg.OpenKey(getattr(winreg, root), p)
    if name:
        winreg.DeleteValue(key2, name)
    else:
        DeleteRecu(key2)
        winreg.DeleteKey(getattr(winreg, root), p)


def SetKey(key, name='', val=''):
    root, p = key.split('\\', 1)
    key2 = winreg.CreateKey(getattr(winreg, root), p)
    winreg.SetValueEx(key2, name, 0, winreg.REG_SZ, val)


def right_py():
    SetKey(r'HKEY_CURRENT_USER\Software\Classes\.py\ShellNew', 'FileName')


__all__ = [k for k in globals() if k not in __all__]



"""Useful Functions with Built-in Library"""

import os
import html
import time
import traceback
import urllib.parse
import urllib.request

# LOG = time.strftime('RECORD_%Y%m%d_%H%M%S.log')
LOG = time.strftime('LOG_%Y%m%d_%H%M%S.txt')


__all__ = list(globals())


# ---------------------------------------------------------------------------
# Debug
# ---------------------------------------------------------------------------


def tag():
    return time.strftime('[%Y-%m-%d %H:%M:%S] ')


def log(*value):
    s = tag() + ' '.join(map(str, value))
    with open(LOG, 'a', encoding='u8') as f:
        f.write(s)


def check(obj, patt='.*'):
    import re
    import sys
    import inspect
    patt = re.compile(patt)
    print('\nobj:', obj)
    for key in sorted(dir(obj)):
        attr = getattr(obj, key)
        try:
            key += str(inspect.signature(attr))
        except (TypeError, ValueError):
            pass
        if patt.fullmatch(key):
            key = '.' + key
            if not callable(attr):
                result = repr(attr)
                key += ' = '
            else:
                try:
                    result = repr(attr())
                    key += ' = '
                except Exception:
                    result = ''
            print(key, end='')
            print(result, file=sys.stderr)


fps_n = -1
fps_t1 = 0
def fps():
    import time
    global fps_n, fps_t1
    fps_n += 1
    fps_t2 = time.time()
    if fps_t1 == 0:
        fps_t1 = fps_t2
    if fps_t2 - fps_t1 > 1:
        print('fps: %.1f' % (fps_n / (fps_t2 - fps_t1)))
        fps_t1 = fps_t2
        fps_n = 0
    return True


# ---------------------------------------------------------------------------
# File system
# ---------------------------------------------------------------------------


stem = lambda p: os.path.splitext(os.path.basename(p))[0]
select = lambda path: os.popen('explorer /select, "%s"' % os.path.abspath(path))


def path_mark(path, mark='.bak'):
    root, ext = os.path.splitext(path)
    return root + mark + ext


def path_quote(p, repl=None):  # not include path
    for c in '\r\n\t\\/:*?"<>|':
        new = repl if repl is not None else urllib.parse.quote_plus(c)
        p = p.replace(c, new)
    return p


def path_split(p):
    root, file = os.path.split(p)
    name, ext = os.path.splitext(file)
    return root, name, ext


def path_unique(p, dash='-'):
    root, ext = os.path.splitext(p)
    n = 0
    while os.path.exists(p):
        n += 1
        p = '%s%s%d%s' % (root, dash, n, ext)
    return p


# ---------------------------------------------------------------------------
# Web
# ---------------------------------------------------------------------------


escape = html.escape
unescape = html.unescape
quote = urllib.parse.quote_plus  # quote every word include '/'
unquote = urllib.parse.unquote


def urlopen(url, timeout=5):
    request = urllib.request.Request(url)
    response = urllib.request.urlopen(request, timeout=timeout)
    return response.read()


# ---------------------------------------------------------------------------
# ...
# ---------------------------------------------------------------------------


def findpair(s, p1='(', p2=')', st=0):
    n1 = n2 = 0
    for n, c in enumerate(s[st:]):
        n1 += c in p1
        n2 += c in p2
        if n1 and n1 == n2:
            return st + n


def bytes_format(data, n=16):
    '''TEST: bytes_format(bytes(range(128)))'''
    s = ''.join('\\x%02x' % c for c in data)
    return "(b'%s')" % "'\n b'".join(s[i:i+n*4] for i in range(0, len(s), n * 4))


def bytes_print(data, n=16):
    print(bytes_format(data, n))


def bytes_hex(data, offset=0, length=-1, slice=-1):
    s = ' '.join('%02X' % c for c in data)
    s = ['%08X ' % i + s[i*3:(i+16)*3] for i in range(0, len(data), 16)]
    s = '\n'.join(s)
    return s


def install(path, block=False):
    p = os.popen('pip install "%s"' % path)
    if block:
        return p.read()


def input_wait(msg):
    while input(msg + '[y/n]: ').lower() != 'y':
        pass


input_default = lambda msg, default: input('input <%s>, keep <%s> press enter: ' % (msg, default)) or default


class Catch:
    def __init__(self, log='log.txt'):
        self.log = log

    def __enter__(self):
        pass

    def __exit__(self, *args):
        if any(args):
            traceback.print_exc()
            error = traceback.format_exc()
            tt = time.strftime('[%Y-%m-%d %H:%M:%S] ')
            with open(self.log, 'a', encoding='u8') as f:
                f.write(tt + error + '\n')
        return True


catch = Catch()


__all__ = [k for k in globals() if k not in __all__]



"""Image Library Functions"""

import os


__all__ = list(globals())


def imread(file):
    import cv2
    import numpy as np
    return cv2.imdecode(np.fromfile(file, np.uint8), -1)


def imwrite(file, im):
    import cv2
    cv2.imencode('.jpg', im)[1].tofile(file)


def imshow(img, delay=50, title=''): # for test
    import cv2
    cv2.namedWindow(title)
    cv2.imshow(title, img)
    cv2.waitKey(delay)


def imsave(file): # for test
    import cv2
    folder = os.path.splitext(file)[0]
    os.makedirs(folder, exist_ok=True)
    for n, img in enumerate(imiter(file)):
        cv2.imwrite('%s/img_%04d.png' % (folder, n), img)


def imiter(file_or_id):
    import cv2
    cap = cv2.VideoCapture(file_or_id)
    while cap.isOpened():
        ret, img = cap.read()
        if not ret:
            cap.release()
            return
        yield img


__all__ = [k for k in globals() if k not in __all__]



import os
import shutil
import setuptools

pkg = setuptools.find_packages()

keys = ['readme', 'license', 'todo', 'version']
os.makedirs(pkg[0] + '/docs', exist_ok=True)
for file in os.listdir():
    if any(file.lower().startswith(k) for k in keys):
        shutil.copy(file, pkg[0] + '/docs/' + file.lower())

with open('VERSION.txt') as f:
    version = f.read().strip()

with open('README.md', encoding='u8') as f:
    long_description = f.read()

setuptools.setup(
    name=pkg[0],
    version=version,
    author='Lishixian(znsoooo)',
    author_email='lsx7@sina.com',
    description='lishixian package',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/znsoooo/lishixian',
    project_urls={
        'Bug Tracker': 'https://github.com/znsoooo/lishixian/issues',
    },
    classifiers=[
        'Programming Language :: Python :: 3',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
    ],
    packages=pkg,

    # install_requires=['pkg_name'],
    # extras_requires={'pkg_name': ['pkg_name']},
    license='MIT License',
    package_data={'': ['*.*']},
    keywords='lishixian lsx',
)



"""Office Automation"""

import os
import sys
import time
from threading import Thread


__all__ = list(globals())


def shortcut(p=None, make=True):  # get_path: desktop/programs/startup/...
    import winshell
    p = p or winshell.desktop()
    path = sys.argv[0]
    name = os.path.splitext(os.path.basename(path))[0]
    target = os.path.join(p, name)
    if make:
        winshell.CreateShortcut(path, target)
    elif os.path.exists(target): # no use
        os.remove(target)


def copy(word, tab=0):
    import pyautogui
    import pyperclip
    for i in range(tab):
        pyautogui.press('tab')
    bak = pyperclip.paste()
    pyperclip.copy(word)
    pyautogui.hotkey('ctrl', 'v')
    pyperclip.copy(bak)


class Monitor:
    def __init__(self, func):
        import pynput
        self.pynput = pynput
        self.func = func
        Thread(target=self.mouser).start()
        Thread(target=self.keyboarder).start()

    def hook(self, *typ_names):
        for name in typ_names:
            yield (lambda typ: (lambda *args: self.func(typ, *args)))(name)

    def mouser(self):
        with self.pynput.mouse.Listener(*self.hook('move', 'click', 'scroll')) as self.ml:
            self.ml.join()

    def keyboarder(self):
        with self.pynput.keyboard.Listener(*self.hook('press', 'release')) as self.kl:
            self.kl.join()

    def stop(self):
        self.pynput.keyboard.Listener.stop(self.kl)
        self.pynput.mouse.Listener.stop(self.ml)


class Recoder:
    def __init__(self, complete=False):
        self.complete = complete

        self.clock = time.time()
        self.monitor = Monitor(self.callback)
        self.stop = self.monitor.stop

    def callback(self, typ, *args):
        if typ == 'move':
            if self.complete:
                x, y = args
                self.log('pyautogui.moveTo(%d, %d)' % (x, y))

        elif typ == 'click':
            x, y, button, pressed = args
            if self.complete:
                event = 'mouseDown' if pressed else 'mouseUp'
                self.log("pyautogui.%s(%d, %d, '%s')" % (event, x, y, button.name))
            elif pressed:
                self.log("pyautogui.click(%d, %d)" % (x, y))

        elif typ == 'scroll':
            x, y, dx, dy = args
            self.log('Scrolled {0}'.format((x, y))) # todo

        elif typ in ('press', 'release'):
            key, = args
            event = 'keyDown' if typ == 'press' else 'keyUp'
            self.log("pyautogui.%s(%s)" % (event, key))

    def log(self, msg):
        dt = time.time() - self.clock
        if self.complete:
            print('time.sleep(%.3f)' % dt)
            self.clock += dt
        elif dt > 0.5:
            dt = dt - dt % 0.5
            print('time.sleep(%.1f)' % dt)
            self.clock += dt
        print(msg)


__all__ = [k for k in globals() if k not in __all__]


if __name__ == '__main__':
    m = Recoder()
    # m = Recoder(complete=True)
    time.sleep(2)
    m.stop()


# todo 全局快捷键（组合键）
